from agrippa.parse import export  # So it's available when you "import agrippa"

from agrippa import utils

from agrippa.onnx2torch import convert as onnx_to_torch

# Import globals
from agrippa.parse import WEIGHTS_FNAME
from agrippa.parse import META_FNAME
from agrippa.parse import ONNX_TYPE_DICT


__all__ = [
    'CustomExportToOnnx',
    'OnnxToTorchModuleWithCustomExport',
]

from typing import Any

import torch
from torch import _C as torch_C

from agrippa.onnx2torch.utils.common import OnnxToTorchModule


class OnnxToTorchModuleWithCustomExport(OnnxToTorchModule):
    """
    Marker class for agrippa.onnx2torch modules with custom export to onnx.
    """

    pass  # pylint: disable=unnecessary-pass


class CustomExportToOnnx(torch.autograd.Function):  # pylint: disable=missing-class-docstring
    _NEXT_OUTPUT = None

    @classmethod
    def set_output_and_apply(cls, output: Any, *args) -> Any:  # pylint: disable=missing-function-docstring
        CustomExportToOnnx._NEXT_OUTPUT = output
        return cls.apply(*args)

    @staticmethod
    def forward(  # pylint: disable=unused-argument, missing-function-docstring
        ctx: Any,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        return CustomExportToOnnx._NEXT_OUTPUT

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:  # pylint: disable=unused-argument, missing-function-docstring
        raise RuntimeError('backward called while converting to onnx')

    @staticmethod
    def symbolic(  # pylint: disable=unused-argument, missing-function-docstring
        graph: torch_C.Graph,
        *values,
    ) -> torch_C.Value:
        raise NotImplementedError



__all__ = [
    'OnnxMinMax',
]

import torch
import torch._C as torch_C

from agrippa.onnx2torch.node_converters.base_element_wise import OnnxBaseElementWise
from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx

_OPERATORS = {
    'Min': torch.amin,
    'Max': torch.amax,
}


class _MinExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('Min', *args, outputs=1)


class _MaxExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('Max', *args, outputs=1)


_CUSTOM_ONNX_EXPORT_CLASS = {
    'Min': _MinExportToOnnx,
    'Max': _MaxExportToOnnx,
}


class OnnxMinMax(OnnxBaseElementWise):  # pylint: disable=missing-docstring
    def __init__(self, operation_type: str):
        super().__init__(_CUSTOM_ONNX_EXPORT_CLASS[operation_type])
        self._operator = _OPERATORS[operation_type]

    def apply_reduction(self, *tensors: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        broadcast_shape = self._broadcast_shape(*tensors)
        broadcast_tensors = [t.broadcast_to(broadcast_shape) for t in tensors]
        stacked_tensors = torch.stack(broadcast_tensors)
        output = self._operator(stacked_tensors, dim=0)
        return output


@add_converter(operation_type='Min', version=8)
@add_converter(operation_type='Min', version=12)
@add_converter(operation_type='Min', version=13)
@add_converter(operation_type='Max', version=8)
@add_converter(operation_type='Max', version=12)
@add_converter(operation_type='Max', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxMinMax(node.operation_type),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



import pickle
import os

"""

NOTE: This will need to change as there are other options for specifying weight files
i.e., in order to support very large weight files

ALSO NOTE: save_pytorch_model depends on the behavior of onnx2torch prepending ".initializers" to every name

"""

# Searches for weights that contain the name arg as a substring in the project
# Args: substring for search, project directory, weights filename (optional)
# Returns dictionary of potential matches with real name -> value
def find_params(name, proj_dir, weights_fname="weights.pkl"):
    weights_file = os.path.join(proj_dir, weights_fname)
    matches = dict()
    with open(weights_file, "rb") as fhand:
        weight_dict = pickle.load(fhand)
        for key in weight_dict:
            if name in key:
                matches[key] = weight_dict[key]
    return matches

# Takes a pytorch model and saves its weights file into the directory specified with the given name (by default, weights.pkl)
def save_torch_model(torch_model, proj_dir, weights_fname="weights.pkl"):
    weights_dict = {}
    whole_dict = torch_model.state_dict()
    for key in whole_dict:
        real_name = key[len('initializers.'):]
        weights_dict[real_name] = whole_dict[key]

    with open(os.path.join(proj_dir, weights_fname), 'wb') as fhand:
        pickle.dump(weights_dict, fhand)



__all__ = [
    'OnnxScatterND',
]

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxScatterND(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def forward(  # pylint: disable=missing-function-docstring
        self,
        data: torch.Tensor,
        indices: torch.Tensor,
        updates: torch.Tensor,
    ) -> torch.Tensor:
        # There is no scatter nd for torch, use following formula:
        # https://github.com/onnx/onnx/blob/master/docs/Operators.md#ScatterND
        output = data.clone()

        if torch.onnx.is_in_onnx_export():
            return _ScatterNDExportToOnnx.set_output_and_apply(output, data, indices, updates)

        ind_dim = indices.dim()
        # last dimension is a partial-index into data
        indices = indices.reshape((-1, indices.shape[-1])).T.tolist()
        # update.shape = indices.shape[0:ind_dim-1] ++ data.shape[indices.shape[-1]:data.dim()-1]
        updates = updates.reshape((-1, *updates.shape[ind_dim - 1 :]))
        output[indices] = updates

        return output


class _ScatterNDExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('ScatterND', *args, outputs=1)


@add_converter(operation_type='ScatterND', version=11)
@add_converter(operation_type='ScatterND', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxScatterND(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxShape',
]

from typing import Optional

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxShape(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def __init__(self, start: Optional[int] = None, end: Optional[int] = None):
        super().__init__()
        self.start = start
        self.end = end

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        output = torch.tensor(
            input_tensor.shape[self.start : self.end],
            device=input_tensor.device,
        )
        if torch.onnx.is_in_onnx_export():
            args = [
                input_tensor,
            ]
            if self.start is not None:
                args.append(self.start)
                if self.end is not None:
                    args.append(self.end)
            elif self.end is not None:
                args += [0, self.end]

            return _ShapeExportToOnnx.set_output_and_apply(output, *args)

        return output


class _ShapeExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        if len(args) == 2:
            return graph.op('Shape', args[0], start_i=args[1], outputs=1)

        if len(args) == 3:
            return graph.op('Shape', args[0], start_i=args[1], end_i=args[2], outputs=1)

        return graph.op('Shape', *args, outputs=1)


@add_converter(operation_type='Shape', version=1)
@add_converter(operation_type='Shape', version=13)
@add_converter(operation_type='Shape', version=15)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxShape(
            start=node.attributes.get('start', None),
            end=node.attributes.get('end', None),
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxTile',
]

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxTile(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        repeats: torch.Tensor,
    ) -> torch.Tensor:
        # torch.tile(input_tensor, repeats) is not supported for exporting
        output = input_tensor.repeat(torch.Size(repeats))
        if torch.onnx.is_in_onnx_export():
            return _TileExportToOnnx.set_output_and_apply(output, input_tensor, repeats)

        return output


class _TileExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('Tile', *args, outputs=1)


@add_converter(operation_type='Tile', version=6)
@add_converter(operation_type='Tile', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument

    return OperationConverterResult(
        torch_module=OnnxTile(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxMatMul',
]

import torch
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OnnxToTorchModule
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node


class OnnxMatMul(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        return torch.matmul(x, y)


@add_converter(operation_type='MatMul', version=1)
@add_converter(operation_type='MatMul', version=9)
@add_converter(operation_type='MatMul', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxMatMul(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxRoiAlign',
]

from typing import Tuple

import torch
import torch._C as torch_C
from torch import nn
from torchvision.ops import roi_align

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxRoiAlign(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def __init__(
        self,
        mode: str = 'avg',
        output_height: int = 1,
        output_width: int = 1,
        sampling_ratio: int = 0,
        spatial_scale: float = 1.0,
    ):
        super().__init__()
        if mode != 'avg':
            raise NotImplementedError(f'"{mode}" roi align mode is not implemented.')

        self._output_height = output_height
        self._output_width = output_width
        self._sampling_ratio = sampling_ratio
        self._spatial_scale = spatial_scale

    @staticmethod
    def _do_forward(
        input_tensor: torch.Tensor,
        rois: torch.Tensor,
        batch_indices: torch.Tensor,
        output_size: Tuple[int, int],
        sampling_ratio: int,
        spatial_scale: float,
    ) -> torch.Tensor:

        batch_indices = batch_indices.unsqueeze(1).to(rois.dtype)
        batched_rois = torch.cat([batch_indices, rois], dim=1)

        return roi_align(
            input=input_tensor,
            boxes=batched_rois,
            output_size=output_size,
            spatial_scale=spatial_scale,
            sampling_ratio=sampling_ratio,
            aligned=False,
        )

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        rois: torch.Tensor,
        batch_indices: torch.Tensor,
    ) -> torch.Tensor:

        output = self._do_forward(
            input_tensor=input_tensor,
            rois=rois,
            batch_indices=batch_indices,
            output_size=(self._output_height, self._output_width),
            sampling_ratio=self._sampling_ratio,
            spatial_scale=self._spatial_scale,
        )
        if torch.onnx.is_in_onnx_export():
            args = [
                input_tensor,
                rois,
                batch_indices,
                self._output_height,
                self._output_width,
                self._sampling_ratio,
                self._spatial_scale,
            ]
            return _RoiAlignExportToOnnx.set_output_and_apply(output, *args)

        return output


class _RoiAlignExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        input_args = args[:3]
        # pylint: disable=unbalanced-tuple-unpacking
        output_height, output_width, sampling_ratio, spatial_scale = args[3:]
        return graph.op(
            'RoiAlign',
            *input_args,
            output_height_i=output_height,
            output_width_i=output_width,
            sampling_ratio_i=sampling_ratio,
            spatial_scale_f=spatial_scale,
            outputs=1,
        )


@add_converter(operation_type='RoiAlign', version=10)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    mode = node_attributes.get('mode', 'avg')
    output_height = node_attributes.get('output_height', 1)
    output_width = node_attributes.get('output_width', 1)
    sampling_ratio = node_attributes.get('sampling_ratio', 0)
    spatial_scale = node_attributes.get('spatial_scale', 1.0)

    return OperationConverterResult(
        torch_module=OnnxRoiAlign(
            mode=mode,
            output_height=output_height,
            output_width=output_width,
            sampling_ratio=sampling_ratio,
            spatial_scale=spatial_scale,
        ),
        onnx_mapping=onnx_mapping_from_node(node),
    )



__all__ = [
    'OnnxEinsum',
]

import torch
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OnnxToTorchModule
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node


class OnnxEinsum(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, equation: str):
        super().__init__()
        self.equation = equation

    def forward(self, *args):  # pylint: disable=missing-function-docstring
        return torch.einsum(self.equation, *args)


@add_converter(operation_type='Einsum', version=12)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxEinsum(
            equation=node.attributes['equation'],
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxGlobalAveragePool',
    'OnnxGlobalAveragePoolWithKnownInputShape',
]

from typing import List

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import get_shape_from_value_info
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxGlobalAveragePool(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-docstring
    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        x_dims = list(range(2, len(input_tensor.shape)))
        output = torch.mean(input_tensor, dim=x_dims, keepdim=True)
        if torch.onnx.is_in_onnx_export():
            return _GlobalAveragePoolExportToOnnx.set_output_and_apply(output, input_tensor)

        return output


class OnnxGlobalAveragePoolWithKnownInputShape(
    nn.Module, OnnxToTorchModuleWithCustomExport
):  # pylint: disable=missing-docstring
    def __init__(self, input_shape: List[int]):
        super().__init__()
        self.x_dims = list(range(2, len(input_shape)))

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        output = torch.mean(input_tensor, dim=self.x_dims, keepdim=True)
        if torch.onnx.is_in_onnx_export():
            return _GlobalAveragePoolExportToOnnx.set_output_and_apply(output, input_tensor)

        return output


class _GlobalAveragePoolExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('GlobalAveragePool', *args, outputs=1)


@add_converter(operation_type='GlobalAveragePool', version=1)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    input_value_info = graph.value_info[node.input_values[0]]
    input_shape = get_shape_from_value_info(input_value_info)

    if input_shape is not None:
        torch_module = OnnxGlobalAveragePoolWithKnownInputShape(input_shape=input_shape)
    else:
        torch_module = OnnxGlobalAveragePool()

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxGather',
    'OnnxGatherElements',
]

from typing import List
from typing import Tuple
from typing import Union

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OnnxToTorchModule
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxGatherElements(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, axis: int = 0):
        super().__init__()
        self.axis = axis

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        indices: torch.Tensor,
    ) -> torch.Tensor:
        return torch.gather(input_tensor, dim=self.axis, index=indices)


class OnnxGather(nn.Module, OnnxToTorchModuleWithCustomExport):
    """ONNX gather implementation (or numpy.take implementation)"""

    def __init__(self, axis: int = 0):
        super().__init__()
        self.axis = axis

    @staticmethod
    def slice_from_axis(  # pylint: disable=missing-docstring
        input_tensor: torch.Tensor,
        axis: int,
        indices: torch.Tensor,
    ) -> Tuple[Union[slice, torch.Tensor], ...]:
        axis = input_tensor.dim() + axis if axis < 0 else axis
        skip_axis: List[Union[slice, torch.Tensor]] = [slice(None)] * axis
        skip_axis.append(indices)
        return tuple(skip_axis)

    def forward(  # pylint: disable=missing-function-docstring
        self, input_tensor: torch.Tensor, indices: torch.Tensor
    ) -> torch.Tensor:
        # pytorch Gather differs from onnx Gather, onnx gather work like numpy.take
        # But torch.take does not support different axis. So we make it by yourself
        # numpy.take is input_data[:, :, indices] where we pass NONE slices AXIS time
        slice_for_take = self.slice_from_axis(input_tensor, self.axis, indices)
        output = input_tensor[slice_for_take]
        if torch.onnx.is_in_onnx_export():
            return _GatherExportToOnnx.set_output_and_apply(output, input_tensor, indices, self.axis)

        return output


class _GatherExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        input_tensor, indices, axis = args
        return graph.op('Gather', input_tensor, indices, axis_i=axis, outputs=1)


@add_converter(operation_type='Gather', version=1)
@add_converter(operation_type='Gather', version=11)
@add_converter(operation_type='Gather', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axis = node.attributes.get('axis', 0)
    torch_module = OnnxGather(
        axis=axis,
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='GatherElements', version=11)
@add_converter(operation_type='GatherElements', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axis = node.attributes.get('axis', 0)
    torch_module = OnnxGatherElements(
        axis=axis,
    )

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



import json
from lib2to3.pgen2 import token

# OP: PRECEDENCE
BIN_OP_PRECEDENCE = {
    '^': 2,
    '*': 1,
    '/': 1,
    '//': 1,
    '%': 1,
    '-': 0,
    '+': 0,
}

BIN_OP_FUNCTIONS = {
    '^': lambda x, y: x**y,
    '*': lambda x, y: x*y,
    '/': lambda x, y: x/y,
    '//': lambda x, y: x // y,
    '%': lambda x, y: x%y,
    '-': lambda x, y: x-y,
    '+': lambda x, y: x+y
}

# For dealing with two char long ops
# suffix -> full op
BIN_OP_SUFFIXES = {
    '/': '//'
}

def _tokenize(text, bindings):
    tokens = []

    working_text = ""
    for c in text:
        if c in BIN_OP_PRECEDENCE:
            if working_text:
                tokens.append(working_text)
                working_text = ""
            if c in BIN_OP_SUFFIXES:
                if len(tokens) > 0:
                    working_op = tokens[-1] + c
                    if working_op in BIN_OP_PRECEDENCE:
                        tokens[-1] = working_op
                        continue
            tokens.append(c)
        elif c in [' ']:  # separation characters
            if working_text:
                tokens.append(working_text)
                working_text = ""
        else:
            working_text += c
    if working_text:
        tokens.append(working_text)

    # Apply bindings
    tokens = [str(bindings[x]) if x in bindings else x for x in tokens]
    return tokens

def _has_leq_precedence(first, second):
    return BIN_OP_PRECEDENCE[first] <= BIN_OP_PRECEDENCE[second]

def _infix_to_prefix(infix):
    operators = []
    operands = []
 
    for i in range(len(infix)):
        if infix[i] not in BIN_OP_FUNCTIONS:
            operands.append(infix[i])
        else:
            while len(operators)!=0 and _has_leq_precedence(infix[i], operators[-1]):
                op1 = operands[-1]
                operands.pop()
 
                op2 = operands[-1]
                operands.pop()
 
                op = operators[-1]
                operators.pop()
 
                tmp = [op] + [op2] + [op1]
                operands.append(tmp)
            operators.append(infix[i])
 
    while (len(operators)!=0):
        op1 = operands[-1]
        operands.pop()
 
        op2 = operands[-1]
        operands.pop()
 
        op = operators[-1]
        operators.pop()
 
        tmp = [op] + [op2] + [op1]
        operands.append(tmp)

    return operands[-1]

def _eval(toks):

    # Sometimes this function gets a singleton expression (string)
    if type(toks) != type([]):
        return json.loads(toks)

    toks = toks[::-1]
    stack = []
    for tok in toks:
        if tok in BIN_OP_PRECEDENCE:
            operand1 = stack.pop()
            operand2 = stack.pop()
            # If we're operating on them, they must be values of some kind
            # Kind of hacky
            operand1 = json.loads(str(operand1))
            operand2 = json.loads(str(operand2))

            func = BIN_OP_FUNCTIONS[tok]
            try:
                app = func(operand1, operand2)
            except:
                raise SyntaxError(f"Incompatible operands {operand1}, {operand2} for operator {tok}")
            stack.append(app)
        else:
            stack.append(tok)
    if len(stack) == 0:
        raise SyntaxError("Invalid expression")
    return stack[0]

def parse_expr(text, bindings={}):

    if '(' in text or ')' in text:
        raise SyntaxWarning("Parentheses not supported; ignoring.")

    tokens = _tokenize(text, bindings)

    prefixed = _infix_to_prefix(tokens)

    evaluated = _eval(prefixed)

    return evaluated



__all__ = [
    'OnnxExpand',
]

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxExpand(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-docstring
    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        shape: torch.Tensor,
    ) -> torch.Tensor:
        output = input_tensor * torch.ones(torch.Size(shape), dtype=input_tensor.dtype, device=input_tensor.device)
        if torch.onnx.is_in_onnx_export():
            return _ExpandExportToOnnx.set_output_and_apply(output, input_tensor, shape)

        return output


class _ExpandExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('Expand', *args, outputs=1)


@add_converter(operation_type='Expand', version=8)
@add_converter(operation_type='Expand', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxExpand(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxUnsqueezeStaticAxes',
    'OnnxUnsqueezeDynamicAxes',
]

from typing import List

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OnnxMapping
from agrippa.onnx2torch.utils.common import OnnxToTorchModule
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import get_const_value
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxUnsqueezeStaticAxes(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, axes: List[int]):
        super().__init__()
        self.axes = sorted(axes)

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        result = input_tensor
        for axes_id in self.axes:
            result = torch.unsqueeze(result, dim=axes_id)

        return result


class OnnxUnsqueezeDynamicAxes(  # pylint: disable=missing-class-docstring
    nn.Module,
    OnnxToTorchModuleWithCustomExport,
):
    @staticmethod
    def _do_forward(input_tensor: torch.Tensor, axes: torch.Tensor) -> torch.Tensor:
        result = input_tensor
        for axes_id in torch.sort(axes).values:
            result = torch.unsqueeze(result, dim=axes_id)

        return result

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        axes: torch.Tensor,
    ) -> torch.Tensor:
        output = self._do_forward(input_tensor, axes)
        if torch.onnx.is_in_onnx_export():
            return _UnsqueezeExportToOnnx.set_output_and_apply(output, input_tensor, axes)

        return output


class _UnsqueezeExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('Unsqueeze', *args, outputs=1)


@add_converter(operation_type='Unsqueeze', version=1)
@add_converter(operation_type='Unsqueeze', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    axes = node.attributes['axes']
    return OperationConverterResult(
        torch_module=OnnxUnsqueezeStaticAxes(axes=axes),
        onnx_mapping=onnx_mapping_from_node(node),
    )


@add_converter(operation_type='Unsqueeze', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    try:
        axes = get_const_value(node.input_values[1], graph)
        axes = axes.tolist()
        return OperationConverterResult(
            torch_module=OnnxUnsqueezeStaticAxes(axes=axes),
            onnx_mapping=OnnxMapping(
                inputs=(node.input_values[0],),
                outputs=node.output_values,
            ),
        )
    except KeyError:
        pass

    return OperationConverterResult(
        torch_module=OnnxUnsqueezeDynamicAxes(),
        onnx_mapping=onnx_mapping_from_node(node),
    )



import torch
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OnnxToTorchModule
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node


class OnnxLpNormalization(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, axis, p):
        super().__init__()
        self.axis = axis
        if not p:
            self.p = -1
        else:
            self.p = p

    def forward(self, x):  # pylint: disable=missing-function-docstring
        # An epsilon value is added for stability; we use the default pytorch epsilon
        return torch.nn.functional.normalize(x, p=self.p, dim=self.axis)


@add_converter(operation_type='LpNormalization', version=1)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxLpNormalization(
            axis=node.attributes['axis'],
            p=node.attributes['p']
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxNot',
    'OnnxLogical',
]

from typing import Optional

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OnnxToTorchModule
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import old_style_broadcast
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport

_TORCH_FUNCTION_FROM_ONNX_TYPE = {
    'Or': torch.logical_or,
    'And': torch.logical_and,
    'Xor': torch.logical_xor,
}


class OnnxNot(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        output = torch.logical_not(input_tensor)
        if torch.onnx.is_in_onnx_export():
            return _NotExportToOnnx.set_output_and_apply(output, input_tensor)

        return output


class _NotExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('Not', *args, outputs=1)


class OnnxLogical(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(self, operation_type: str, broadcast: Optional[int] = None, axis: Optional[int] = None):
        super().__init__()
        self.broadcast = broadcast
        self.axis = axis

        self.logic_op_function = _TORCH_FUNCTION_FROM_ONNX_TYPE[operation_type]

    def forward(  # pylint: disable=missing-function-docstring
        self, first_tensor: torch.Tensor, second_tensor: torch.Tensor
    ):
        if self.broadcast == 1 and self.axis is not None:
            second_tensor = old_style_broadcast(first_tensor, second_tensor, self.axis)
        return self.logic_op_function(first_tensor, second_tensor)


@add_converter(operation_type='Xor', version=1)
@add_converter(operation_type='Xor', version=7)
@add_converter(operation_type='And', version=1)
@add_converter(operation_type='And', version=7)
@add_converter(operation_type='Or', version=1)
@add_converter(operation_type='Or', version=7)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxLogical(
            operation_type=node.operation_type,
            broadcast=node.attributes.get('broadcast', None),
            axis=node.attributes.get('axis', None),
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='Not', version=1)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxNot(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxReshape',
]

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxReshape(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    @staticmethod
    def _do_reshape(input_tensor: torch.Tensor, shape: torch.Tensor) -> torch.Tensor:
        if torch.any(shape == 0):
            shape = [input_tensor.shape[i] if dim_size == 0 else dim_size for i, dim_size in enumerate(shape)]

        return torch.reshape(input_tensor, torch.Size(shape))

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        shape: torch.Tensor,
    ) -> torch.Tensor:
        output = self._do_reshape(input_tensor, shape)

        if torch.onnx.is_in_onnx_export():
            return _ReshapeExportToOnnx.set_output_and_apply(output, input_tensor, shape)

        return output


class _ReshapeExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        return graph.op('Reshape', *args, outputs=1)


@add_converter(operation_type='Reshape', version=5)
@add_converter(operation_type='Reshape', version=13)
@add_converter(operation_type='Reshape', version=14)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    if node.attributes.get('allowzero', 0) == 1:
        raise NotImplementedError('"allowzero=1" is not implemented')

    return OperationConverterResult(
        torch_module=OnnxReshape(),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )



__all__ = [
    'OnnxReduceSumDynamicAxes',
    'OnnxReduceSumStaticAxes',
    'OnnxReduceStaticAxes',
]

from functools import partial
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import torch
import torch._C as torch_C
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OnnxMapping
from agrippa.onnx2torch.utils.common import OnnxToTorchModule
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import get_const_value
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


@torch.fx.wrap
def _get_element(x: Union[List, Tuple], index: int = 0) -> Any:
    if isinstance(x, (tuple, list)):
        return x[index]

    return x


def _initialize_none_dim(dim: Optional[Union[int, Tuple[int, ...]]], input_dim: int):
    if dim is None:
        return list(range(input_dim))

    return dim


def _log_sum(
    input_tensor: torch.Tensor,
    dim: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdim: bool = False,
):
    dim = _initialize_none_dim(dim, input_tensor.dim())
    return torch.log(torch.sum(input_tensor, dim=dim, keepdim=keepdim))


def _log_sum_exp(
    input_tensor: torch.Tensor,
    dim: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdim: bool = False,
):
    dim = _initialize_none_dim(dim, input_tensor.dim())
    return torch.logsumexp(input_tensor, dim=dim, keepdim=keepdim)


def _sum_square(
    input_tensor: torch.Tensor,
    dim: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdim: bool = False,
):
    dim = _initialize_none_dim(dim, input_tensor.dim())
    return torch.sum(torch.square(input_tensor), dim=dim, keepdim=keepdim)


_TORCH_FUNCTION_FROM_ONNX_TYPE = {
    'ReduceL1': partial(torch.norm, p=1),
    'ReduceL2': partial(torch.norm, p=2),
    'ReduceLogSum': _log_sum,
    'ReduceLogSumExp': _log_sum_exp,
    'ReduceMax': torch.max,
    'ReduceMean': torch.mean,
    'ReduceMin': torch.min,
    'ReduceProd': torch.prod,
    'ReduceSum': torch.sum,
    'ReduceSumSquare': _sum_square,
}


class OnnxReduceSumDynamicAxes(  # pylint: disable=missing-class-docstring
    nn.Module,
    OnnxToTorchModuleWithCustomExport,
):
    def __init__(self, keepdims: int = 1, noop_with_empty_axes: int = 0):
        super().__init__()

        self.keepdims = keepdims == 1
        self.noop_with_empty_axes = noop_with_empty_axes == 1

    def _do_forward(self, input_tensor: torch.Tensor, axes: torch.Tensor) -> torch.Tensor:
        if axes is None or axes.nelement() == 0:
            if self.noop_with_empty_axes:
                return input_tensor

            if not self.keepdims:
                return torch.sum(input_tensor)

            axes = list(range(input_tensor.dim()))
        else:
            axes = torch.sort(axes).values.tolist()

        return torch.sum(input_tensor, dim=axes, keepdim=self.keepdims)

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_tensor: torch.Tensor,
        axes: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        output = self._do_forward(input_tensor, axes)
        if torch.onnx.is_in_onnx_export():
            args = [input_tensor]
            if axes is not None:
                args.append(axes)

            return _ReduceSumExportToOnnx.set_output_and_apply(
                output,
                *args,
                int(self.keepdims),
                int(self.noop_with_empty_axes),
            )

        return output


class _ReduceSumExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        *args, keepdims, noop_with_empty_axes = args
        return graph.op(
            'ReduceSum',
            *args,
            noop_with_empty_axes_i=noop_with_empty_axes,
            keepdims_i=keepdims,
            outputs=1,
        )


class OnnxReduceSumStaticAxes(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(
        self,
        axes: List[int],
        keepdims: int = 1,
        noop_with_empty_axes: int = 0,
    ):
        super().__init__()
        if axes is not None:
            axes = sorted(axes)

        self.keepdims = keepdims == 1
        self.noop_with_empty_axes = noop_with_empty_axes == 1
        self.axes = axes

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        if self.axes is None or len(self.axes) == 0:
            if self.noop_with_empty_axes:
                return input_tensor

            if not self.keepdims:
                return self.math_op_function(input_tensor)

            self.axes = list(range(input_tensor.dim()))

        return torch.sum(input_tensor, dim=self.axes, keepdim=self.keepdims)


class OnnxReduceStaticAxes(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-class-docstring
    def __init__(
        self,
        operation_type: str,
        axes: List[int],
        keepdims: int = 1,
    ):
        super().__init__()
        self.operation_type = operation_type
        self.math_op_function = _TORCH_FUNCTION_FROM_ONNX_TYPE[operation_type]

        if axes is not None:
            axes = sorted(axes)

        self.keepdims = keepdims == 1
        self.axes = axes

    def forward(self, input_tensor: torch.Tensor) -> torch.Tensor:  # pylint: disable=missing-function-docstring
        if self.axes is None or len(self.axes) == 0:
            if not self.keepdims:
                return self.math_op_function(input_tensor)

            self.axes = list(range(input_tensor.dim()))

        if self.operation_type not in ['ReduceMax', 'ReduceMin', 'ReduceProd']:
            return self.math_op_function(input_tensor, dim=self.axes, keepdim=self.keepdims)

        result = input_tensor
        for passed_dims, axis in enumerate(self.axes):
            result = self.math_op_function(
                result,
                dim=axis if self.keepdims else axis - passed_dims,
                keepdim=self.keepdims,
            )
            result = _get_element(result, 0)

        return result


@add_converter(operation_type='ReduceL1', version=1)
@add_converter(operation_type='ReduceL1', version=11)
@add_converter(operation_type='ReduceL1', version=13)
@add_converter(operation_type='ReduceL2', version=1)
@add_converter(operation_type='ReduceL2', version=11)
@add_converter(operation_type='ReduceL2', version=13)
@add_converter(operation_type='ReduceLogSum', version=1)
@add_converter(operation_type='ReduceLogSum', version=11)
@add_converter(operation_type='ReduceLogSum', version=13)
@add_converter(operation_type='ReduceLogSumExp', version=1)
@add_converter(operation_type='ReduceLogSumExp', version=11)
@add_converter(operation_type='ReduceLogSumExp', version=13)
@add_converter(operation_type='ReduceMax', version=1)
@add_converter(operation_type='ReduceMax', version=11)
@add_converter(operation_type='ReduceMax', version=12)
@add_converter(operation_type='ReduceMax', version=13)
@add_converter(operation_type='ReduceMean', version=1)
@add_converter(operation_type='ReduceMean', version=11)
@add_converter(operation_type='ReduceMean', version=13)
@add_converter(operation_type='ReduceMin', version=1)
@add_converter(operation_type='ReduceMin', version=11)
@add_converter(operation_type='ReduceMin', version=12)
@add_converter(operation_type='ReduceMin', version=13)
@add_converter(operation_type='ReduceProd', version=1)
@add_converter(operation_type='ReduceProd', version=11)
@add_converter(operation_type='ReduceProd', version=13)
@add_converter(operation_type='ReduceSum', version=1)
@add_converter(operation_type='ReduceSum', version=11)
@add_converter(operation_type='ReduceSumSquare', version=1)
@add_converter(operation_type='ReduceSumSquare', version=11)
@add_converter(operation_type='ReduceSumSquare', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    node_attributes = node.attributes
    axes = node_attributes.get('axes', None)
    keepdims = node_attributes.get('keepdims', 1)

    return OperationConverterResult(
        torch_module=OnnxReduceStaticAxes(
            operation_type=node.operation_type,
            axes=axes,
            keepdims=keepdims,
        ),
        onnx_mapping=onnx_mapping_from_node(node=node),
    )


@add_converter(operation_type='ReduceSum', version=13)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    keepdims = node.attributes.get('keepdims', 1)
    noop_with_empty_axes = node.attributes.get('noop_with_empty_axes', 0)

    if len(node.input_values) == 2:
        try:
            axes = get_const_value(node.input_values[1], graph)
            axes = axes.tolist()
            return OperationConverterResult(
                torch_module=OnnxReduceSumStaticAxes(
                    axes=axes,
                    keepdims=keepdims,
                    noop_with_empty_axes=noop_with_empty_axes,
                ),
                onnx_mapping=OnnxMapping(
                    inputs=(node.input_values[0],),
                    outputs=node.output_values,
                ),
            )
        except KeyError:
            pass

    return OperationConverterResult(
        torch_module=OnnxReduceSumDynamicAxes(keepdims=keepdims, noop_with_empty_axes=noop_with_empty_axes),
        onnx_mapping=onnx_mapping_from_node(node),
    )



from math import exp
from xml.dom import SyntaxErr
import xml.etree.ElementTree as ET
import onnx
import json
import numpy as np
import os
import pickle
from agrippa.expr_parser import parse_expr
import math

WEIGHTS_FNAME = "weights.pkl"
META_FNAME = "meta.json"

ONNX_TYPE_DICT = {
    "float32": onnx.TensorProto.FLOAT
}

DEFAULT_TYPE = "float32"
DEFAULT_INIT_TYPE = "xavier"

suppress_prints = False  # global variable

# For print statements
def _notify(str):
    global suppress_prints
    if not suppress_prints:
        print(str)

"""
Gets raw attribute text, a dictionary of global variable bindings
Returns the true attribute value, with variables and expressions resolved
expect_value is True if the return value is not meant to be a string
"""
def _resolve_attr(text, bindings, expect_value=True):
    if bindings:
        for key in bindings:
            text = text.replace("var("+str(key)+")", str(bindings[key]))
    # detect unbound variable
    if text.find("var(") != -1:
        raise SyntaxWarning(f"Unbound variable found in attribute '{text}'")
    
    # If there's an expression, parse it
    to_replace = {}
    total_offset = 0
    remaining = text[total_offset:]
    while remaining.find('expr(') != -1:
        begin_call = text[total_offset:].find('expr(')
        start_i = begin_call + len("expr(")
        
        expr = ""
        end_place = -1
        parens = []
        for i, c in enumerate(remaining[start_i:]):
            if c == ')':
                if len(parens) == 0:
                    end_place = i + start_i
                    break
                else:
                    parens.pop()
            elif c == '(':
                parens.append('(')
            else:
                expr += c
        
        if end_place == -1:
            raise SyntaxError(f"Mismatched parenthesis in expression '{text}'")
        if not bindings:
            bindings = {}

        parsed_expr = parse_expr(expr, bindings=bindings)
        end_call = end_place + len(")")

        to_replace[remaining[begin_call:end_call]] = str(parsed_expr)

        total_offset += end_call
        remaining = text[total_offset:]

    # Substitute the expressions for their values
    for key in to_replace:
        text = text.replace(key, to_replace[key])

    if expect_value:
        return json.loads(text)
    return text


"""
Searches in the weights file for a corresponding weight value or creates a new initializer
Options for init_type are:
- normal, uni_random, ones, zeros
- init_args takes parameterizations of those initialization types
See the README for more info
"""
def _resolve_param(name, data_type, dims, weights, init_type=DEFAULT_INIT_TYPE, init_args=None):
    if name in weights:
        try:
            tens = weights[name]
            res = onnx.helper.make_tensor(
                name=name,
                data_type=data_type,
                dims=dims,
                vals=tens.flatten().tolist())
            return res
        except:
            _notify(f"Error loading weight {name}; arbitrarily initializing.")

    _notify(f"Weight {name} not found; initializing")
    if init_type == 'xavier':
        # Get the sum of the last two dimensions
        # If there is only one, use just that one
        stdev = 1
        if len(dims) >= 2:
            stdev = math.sqrt(2. / (dims[-1] + dims[-2]))
        elif len(dims) == 1:
            stdev = 1. * math.sqrt(dims[0])
        tens = np.random.normal(loc=0., scale=stdev, size=dims)
    elif init_type == 'normal':
        mu = 0
        sigma = 1
        if init_args is not None:
            if len(init_args) < 2:
                _notify(f"Init args for normal initialization has too few members ({len(init_args)} < 2); using mu=0, std=1")
            else:
                mu = init_args[0]
                sigma = init_args[1]
        tens = np.random.normal(loc=mu, scale=sigma, size=dims)
    elif init_type == 'uni_random':
        a = -1
        b = 1
        if init_args is not None:
            if len(init_args) < 2:
                _notify(f"Init args for uniform random initialization has too few members ({len(init_args)} < 2); using a=-1, b=1")
            else:
                a = init_args[0]
                b = init_args[1]
        tens = np.random.random(dims) * (b - a) + a
    elif init_type == 'ones':
        tens = np.ones(dims)
    elif init_type == 'zeros':
        tens = np.zeros(dims)
    elif init_type == 'constant':
        if len(init_args) != 1:
            raise SyntaxError(f"For init type = constant, init args must have exactly one member")
        val = init_args[0]
        tens = np.full(dims, val)
    else:
        _notify(f"Weight initialization type '{init_type}' not known; defaulting to N(0, 1)")
        tens = np.random.normal(loc=0., scale=1., size=dims)
    res = onnx.helper.make_tensor(
            name=name,
            data_type=data_type,
            dims=dims,
            vals=tens.flatten().tolist())

    weights[name] = tens  # weights is an object, so this is allowed
    return res

# Determines the output file name of the onnx model
# outfile = what the user wants it named
# infile = the project folder name, which is used as a default (with .onnx)
def _get_output_model_name(outfile, infile):
    if outfile:
        return outfile
    return infile + ".onnx"


# Returns true is block is imported
def _detect_import(block):
    is_imp = 'src' in block.attrib
    if is_imp and 'name' not in block.attrib:
        src = block.attrib['src']
        raise SyntaxError(f"An imported block {src} is missing 'name' attribute")
    return is_imp


def _find_import(block_src, prefix=None):
    true_src = block_src if prefix is None else os.path.join(prefix, block_src)
    imp_tree = ET.parse(true_src)  # should throw error if not well-formed XML
    imp_block = imp_tree.getroot()
    return imp_block

"""
def export(infile, ...)

This function is used to convert a project directory into an ONNX file.
The infile parameter should be a directory.
The outfile parameter defaults to your project name with a .onnx extension.

"""
def export(
        infile,
        outfile=None,
        producer="Unknown",
        graph_name="Unknown",
        write_weights=True,
        suppress=False,
        reinit=False,
        bindings=None,
        index=None  # Main file that things are imported into
    ):

    global suppress_prints
    suppress_prints = suppress

    proj_dir = os.listdir(infile)
    arch_file = None
    if index is None:
        for fil in proj_dir:
            try:
                if fil[-4:] == ".xml" or fil[-4:] == ".agr":
                    arch_file = fil
                    break
            except:
                pass
    else:
        arch_file = index

    if not arch_file:
        raise FileNotFoundError("No model architecture file found. A file with .agr or .xml extension expected.")

    # Look for a meta file
    meta_file = None
    for fil in proj_dir:
        try:
            if fil == META_FNAME:
                meta_file = fil
                break
        except:
            pass
    if not meta_file:
        _notify("No meta file found - using default metadata.")
    else:
        with open(os.path.join(infile, meta_file)) as fhand:
            meta = json.load(fhand)
            try:
                producer = meta['producer']
            except KeyError:
                _notify("Meta file does not contain producer - using default.")
            try:
                graph_name = meta['graph_name']
            except KeyError:
                _notify("Meta file does not contain graph name - using default.")

    # Look for a weights file
    weights = {}
    weights_file = None
    for fil in proj_dir:
        try:
            if fil == WEIGHTS_FNAME:
                weights_file = fil
                break
        except:
            pass
    if not weights_file:
        _notify("No weights file found - using arbitrary initialization.")
    elif reinit:
        _notify("Re-initializing weights, as per your request.")
    else:
        with open(os.path.join(infile, weights_file), "rb") as fhand:
            weights = pickle.load(fhand)

    tree = ET.parse(os.path.join(infile, arch_file))  # should throw error if not well-formed XML
    root = tree.getroot()

    # Get root imports/exports

    all_nodes = []
    all_inputs = []
    all_outputs = []
    all_inits = []

    # Root model imports
    for child in root.findall("import"):

        child_type = DEFAULT_TYPE
        if 'type' in child.attrib:
            child_type = child.attrib['type']

        name = _resolve_attr(child.attrib['from'], bindings, expect_value=False)
        child_type = _resolve_attr(child_type, bindings, expect_value=False)
        dim = _resolve_attr(child.attrib['dim'], bindings, expect_value=True)
        x = onnx.helper.make_tensor_value_info(name,
                                            ONNX_TYPE_DICT[child_type],
                                            dim)
        all_inputs.append(x)

    parameter_repeats = {}  # for shared params, decide whether to create new weight value for a parameter
    def get_unique_param_name(name):
        if name in parameter_repeats:
            parameter_repeats[name] += 1
            return name + "$" + str(parameter_repeats[name])
        parameter_repeats[name] = 1
        return name + "$1"

    # used to make sure we get a unique weight name + we can handle nested repeats
    # must return a different name from name even if it doesn't exist yet in the repeats dictionary
    repeats = {}
    def get_unique_name(name):
        if name in repeats:
            repeats[name] += 1
            return name + "$" + str(repeats[name])
        repeats[name] = 1
        return name + "$1"

    repeat_node_names = {}
    def get_unique_node_name(name):
        if name in repeat_node_names:
            repeat_node_names[name] += 1
            return name + "$" + str(repeat_node_names[name])
        repeat_node_names[name] = 1
        return name + "$1"

    block_id_tracker = {'curr': 0}  # A hack to not technically use a global var
    def make_unique_block_id():
        block_id_tracker['curr'] += 1
        return block_id_tracker['curr']

    def make_imported_name(name, imp_name):
        if imp_name is not None:
            return imp_name + "$" + name
        return name

    # the reason why we made those block ids
    saved_import_names = {}  # block id -> dict of names to restore

    name_resolves = {}  # an important data structure for dealing with repetitions and the resulting name changes

    saved_to_concat_exports = {}  # block id -> dict orig export name -> lists of export names to concat

    # How many programming sins can we commit in the shortest amount of code?
    # A hellish mix of state and functional programming.
    # "imp_name" means import name = i.e., name needed to resolve names in another file
    def parse_block(block, block_id, imp_name=None, imp_path=None, rep_index=0, stretch_index=0):

        # Go through each child of the block
        # Could be block or node
        for node in block.findall('*'):

            # Processing children in correct order w.r.t. nodes so that topological sort works
            if node.tag == 'block' or node.tag == 'model':
                # If it's imported, need to change imp_name and continue
                if _detect_import(node):
                    new_imp_name = make_imported_name(node.attrib['name'], imp_name)
                    imp_block = _find_import(node.attrib['src'], os.path.join(infile, imp_path) if imp_path is not None else infile)
                    new_imp_path = os.path.split(os.path.join(imp_path, node.attrib['src']) if imp_path is not None else node.attrib['src'])[0]
                    parse_block(imp_block, make_unique_block_id(), imp_name=new_imp_name, imp_path=new_imp_path)
                else:
                    parse_block(node, make_unique_block_id(), imp_name=imp_name, imp_path=imp_path)
                continue
            elif node.tag != 'node':
                continue

            try:
                op = _resolve_attr(node.attrib['op'], bindings, expect_value=False)
            except KeyError:
                raise SyntaxError("Op type not specified for some node")

            try:
                title = node.attrib['title']
                title = _resolve_attr(node.attrib['title'], bindings, expect_value=False)
            except KeyError:
                title = "GenericNode"  # node title is optional
            title = make_imported_name(title, imp_name)
            title = get_unique_node_name(title)
            
            # Get inputs and outputs
            input_els = node.findall("input")
            naive_inputs = [_resolve_attr(el.attrib['src'], bindings, expect_value=False) for el in input_els]
            inputs = [make_imported_name(x, imp_name) for x in naive_inputs]  # change name if file is from other source
            inputs = [name_resolves[x] if x in name_resolves else x for x in inputs]

            output_els = node.findall("output")
            naive_outputs = [_resolve_attr(el.attrib['name'], bindings, expect_value=False) for el in output_els]
            outputs = [make_imported_name(x, imp_name) for x in naive_outputs]  # change name if file is from other source
            outputs = [name_resolves[x] if x in name_resolves else x for x in outputs]

            # Make the parameters
            params = []
            param_precedence = []
            all_children = node.findall("*")
            precendence_counter = 0
            for el in all_children:
                if el.tag == "params":
                    param = el
                    param_precedence.append(precendence_counter)
                    precendence_counter += 1
                elif el.tag == "input":
                    precendence_counter += 1
                    continue
                else:
                    continue
                # Note: order can be important!
                # Each op type specifies the order in which inputs are meant to be included
                orig_name = _resolve_attr(param.attrib['name'], bindings, expect_value=False)
                if imp_name is not None:
                    orig_name = make_imported_name(orig_name, imp_name)
                name = orig_name

                try:
                    shared = _resolve_attr(param.attrib['shared'], bindings, expect_value=False)
                except:
                    shared = "no"
                
                dim = _resolve_attr(param.attrib['dim'], bindings, expect_value=True)
                
                param_type = DEFAULT_TYPE
                if 'type' in param.attrib:
                    param_type = param.attrib['type']

                dtype = _resolve_attr(param_type, bindings, expect_value=False)

                try:
                    init_type = _resolve_attr(param.attrib['init'], bindings, expect_value=False)
                except:
                    init_type = DEFAULT_INIT_TYPE
                try:
                    init_args = _resolve_attr(param.attrib['init_args'], bindings, expect_value=True)
                except:
                    init_args = None

                orig_name = _resolve_attr(param.attrib['name'], bindings, expect_value=False)

                # If the parameter is frozen, add "$constant" to the name
                try:
                    frozen = _resolve_attr(param.attrib['frozen'], bindings, expect_value=False)
                except:
                    frozen = "no"
                if frozen == "yes":
                    orig_name += "$constant"

                name = orig_name
                
                if shared == "no":
                    name = get_unique_param_name(orig_name)
                    param_onnx = _resolve_param(name, ONNX_TYPE_DICT[dtype], dim, weights, init_type=init_type, init_args=init_args)
                    all_inits.append(param_onnx)
                else:
                    # First time we see it, even if we keep the name, need to initialize
                    if name not in parameter_repeats:
                        param_onnx = _resolve_param(name, ONNX_TYPE_DICT[dtype], dim, weights, init_type=init_type, init_args=init_args)
                        all_inits.append(param_onnx)
                        parameter_repeats[name] = 1  # so we don't do this next time
                params.append(name)

            kwargs = {}

            def splice_inputs(inputs, params, param_precedence):
                to_return = [None for _ in range(len(inputs) + len(params))]
                input_index = 0
                param_index = 0
                for i in range(len(to_return)):
                    if param_index < len(params) and param_precedence[param_index] == i:
                        to_return[i] = params[param_index]
                        param_index += 1
                    elif input_index < len(inputs):
                        to_return[i] = inputs[input_index]
                        input_index += 1
                return to_return
            # Sometimes there are other tags defining certain attributes
            # Here is where we would need to support new op types
            if op == 'MatMul':
                inputs = splice_inputs(inputs, params, param_precedence)
            elif op == "Add":
                inputs = splice_inputs(inputs, params, param_precedence)
            elif op == "Relu":
                inputs = splice_inputs(inputs, params, param_precedence)
            elif op == "LpNormalization":
                kwargs["axis"] = _resolve_attr(node.attrib['axis'], bindings, expect_value=True)
                kwargs["p"] = _resolve_attr(node.attrib['p'], bindings, expect_value=True)
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) > 1:
                    raise SyntaxWarning(f"LpNormalization should only have one input, yet it has {len(inputs)}")
            elif op == "Transpose":
                try:
                    perm = node.attrib['perm']
                    kwargs["perm"] = _resolve_attr(perm, bindings, expect_value=True)
                except KeyError:
                    pass  # Default perm value, which is reverse
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) > 1:
                    raise SyntaxWarning(f"Transpose should only have one input, yet it has {len(inputs)}")
            elif op == "Identity":
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) > 1:
                    raise SyntaxWarning(f"Identity should only have one input, yet it has {len(inputs)}")
            elif op =="Softmax":
                kwargs["axis"] = _resolve_attr(node.attrib['axis'], bindings, expect_value=True)
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) > 1:
                    raise SyntaxWarning(f"Softmax should only have one input, yet it has {len(inputs)}")
            elif op == "Div":
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) != 2:
                    raise SyntaxWarning(f"Div should only have exactly 2 inputs, yet it has {len(inputs)}")
            elif op == "Mul":
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) != 2:
                    raise SyntaxWarning(f"Mul should have exactly 2 inputs, yet it has {len(inputs)}")
            elif op == "ReduceMean":
                kwargs["axes"] = _resolve_attr(node.attrib['axes'], bindings, expect_value=True)
                try:
                    kwargs["keepdims"] = _resolve_attr(node.attrib['keepdims'], bindings, expect_value=True)
                except KeyError:
                    # Default is 1 (used to be a notify - don't put a notify here, it's jank)
                    pass
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) > 1:
                    raise SyntaxWarning(f"Softmax should only have one input, yet it has {len(inputs)}")
            elif op == "Sub":
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) != 2:
                    raise SyntaxWarning(f"Sub should have exactly 2 inputs, yet it has {len(inputs)}")
            elif op == "Sqrt":
                inputs = splice_inputs(inputs, params, param_precedence)
                if len(inputs) != 1:
                    raise SyntaxWarning(f"Sqrt should only have 1 input, yet it has {len(inputs)}")
            elif op == "Concat":
                inputs = splice_inputs(inputs, params, param_precedence)
                try:
                    kwargs["axis"] = _resolve_attr(node.attrib['axis'], bindings, expect_value=True)
                except KeyError:
                    raise SyntaxError(f"Expecting attribute 'axis' for Concat node")
            elif op == "LeakyRelu":
                inputs = splice_inputs(inputs, params, param_precedence)
                try:
                    kwargs["alpha"] = _resolve_attr(node.attrib['alpha'], bindings, expect_value=True)
                except KeyError:
                    raise SyntaxError(f"Expecting attribute 'alpha' for LeakyRelu node")
            elif op == "Conv":
                # refer to: https://github.com/onnx/onnx/blob/main/docs/Operators.md#Conv
                inputs = splice_inputs(inputs, params, param_precedence)
                try:
                    kwargs["kernel_shape"] = _resolve_attr(node.attrib['kernel_shape'], bindings, expect_value=True)
                except KeyError:
                    pass
                try:
                    kwargs["pads"] = _resolve_attr(node.attrib['pads'], bindings, expect_value=True)
                except KeyError:
                    pass
                # optionals
                try:
                    kwargs["dilations"] = _resolve_attr(node.attrib['dilations'], bindings, expect_value=True)
                except KeyError:
                    pass
                try:
                    kwargs["strides"] = _resolve_attr(node.attrib['strides'], bindings, expect_value=True)
                except KeyError:
                    pass
            else:
                raise SyntaxError(f"Unknown op type '{op}'") 

            new_node = onnx.helper.make_node(
                name=title,
                op_type=op,
                inputs=inputs,
                outputs=outputs,
                **kwargs
            )
            all_nodes.append(new_node)
        
        
        """
        Stretch:
        
        Change intermediate node output names, but keep
            import names
        Save new export names for later
        Call parse block until stretch_index runs out
        If we're the final one, create a new node that concats all the saved export names, and
            this node should have output names associated with the original exports
        Can't stretch with rep, so just return from function

        """

        try:
            stretch = _resolve_attr(block.attrib['stretch'], bindings)
        except KeyError:
            stretch = 1
        
        if stretch > 1:

            curr_exports = []

            # MIGHT NEED TO CHANGE SO I CAN ACCESS INTERMEDIATE NAMES

            # save the export resolves
            for exp in block.findall('export'):
                name = _resolve_attr(exp.attrib['from'], bindings, expect_value=False)
                if imp_name is not None:
                    name = make_imported_name(name, imp_name)
                if block_id in saved_to_concat_exports:
                    try:
                        saved_to_concat_exports[block_id][name].append(name_resolves[name])
                    except KeyError:
                        saved_to_concat_exports[block_id][name].append(name)
                else:
                    try:
                        saved_to_concat_exports[block_id] = {name: [name_resolves[name]]}
                    except KeyError:
                        saved_to_concat_exports[block_id] = {name: [name]}

            if stretch_index < stretch - 1:
                # change the intermediate node resolves
                # Note that iter gets nested children as well
                for node in block.iter('node'):
                    for out_el in node.findall("output"):
                        name = _resolve_attr(out_el.attrib['name'], bindings, expect_value=False)
                        if imp_name is not None:
                            name = make_imported_name(name, imp_name)
                        uni = get_unique_name(name)
                        name_resolves[name] = uni  # for next rep
                stretch_index += 1
                parse_block(block, block_id, imp_name=imp_name, rep_index=rep_index, stretch_index=stretch_index, imp_path=imp_path)
            else:
                # Create concat nodes
                for orig_name in saved_to_concat_exports[block_id]:                
                    concat_inputs = saved_to_concat_exports[block_id][orig_name]
                    new_name = orig_name + "$concat"
                    new_name = get_unique_name(new_name)
                    kwargs = {"axis": -1}  # will probably have to change, or provide options
                    concat_node = onnx.helper.make_node(
                        name=make_imported_name(get_unique_node_name("stretch_concat"), imp_name),
                        op_type="Concat",
                        inputs=concat_inputs,
                        outputs=[new_name],
                        **kwargs
                    )
                    all_nodes.append(concat_node)
                    name_resolves[orig_name] = new_name

        # Forget reps if we're stretching
        if stretch > 1:
            return

        """
        
        Now dealing with reps

        """

        # Does the block have a rep?
        try:
            rep = _resolve_attr(block.attrib['rep'], bindings)
        except KeyError:
            rep = 1

        # Save the name state of your imports so that you can return them when you leave
        if rep_index == 0:
            saved_import_names[block_id] = {}
            for imp in block.findall('import'):
                name = imp.attrib['from']
                if imp_name is not None:
                    name = make_imported_name(name, imp_name)
                try:
                    saved_import_names[block_id][name] = name_resolves[name]
                except KeyError:
                    saved_import_names[block_id][name] = name
        
        rep_index += 1
        if rep_index < rep:
            curr_exports = []
            # Find the current exports so that we can import them next time
            for exp in block.findall('export'):
                bad_name = _resolve_attr(exp.attrib['from'], bindings, expect_value=False)
                good_name = make_imported_name(bad_name, imp_name)
                if good_name in name_resolves:
                    curr_exports.append(name_resolves[good_name])
                else:
                    curr_exports.append(good_name)

            # Imported blocks need to be treated specially for their intermediate node resolves
            for bl in block.iter('block'):
                if _detect_import(bl):
                    sourced_bl = _find_import(bl.attrib['src'], os.path.join(infile, imp_path) if imp_path is not None else infile)
                    for node in sourced_bl.iter('*'):
                        if node.tag != 'node':
                            continue
                        for out_el in node.findall("output"):
                            bad_name = _resolve_attr(out_el.attrib['name'], bindings, expect_value=False)
                            # The double make imported name is to carry over the current imp name into the next
                            good_name = make_imported_name(bad_name, make_imported_name(bl.attrib['name'], imp_name))
                            uni = get_unique_name(good_name)
                            name_resolves[good_name] = uni  # for next rep

            # change the intermediate node resolves as though they were exports
            # this is OK I think; treating exports separately is still good to line up rep imports/exports
            # Note that iter gets nested children as well
            for node in block.iter('node'):
                for out_el in node.findall("output"):
                    bad_name = _resolve_attr(out_el.attrib['name'], bindings, expect_value=False)
                    good_name = make_imported_name(bad_name, imp_name)
                    uni = get_unique_name(good_name)
                    name_resolves[good_name] = uni  # for next rep
            
            # Order matters!
            # change the import resolves
            for i, imp in enumerate(block.findall('import')):
                try:
                    bad_name = _resolve_attr(imp.attrib['from'], bindings, expect_value=False)
                    good_name = make_imported_name(bad_name, imp_name)
                    name_resolves[good_name] = curr_exports[i]
                except IndexError:  # it's ok if some imports come from somewhere else and aren't repeated (they must be the last ones)
                    break

            parse_block(block, block_id, imp_name=imp_name, rep_index=rep_index, imp_path=imp_path)
        else:
            # Restore import name resolves and leave
            for name in saved_import_names[block_id]:
                name_resolves[name] = saved_import_names[block_id][name]


    # Go through each block (does this recursively)
    for block in root.findall('block'):
        # Is block imported?
        if _detect_import(block):
            imp_name = block.attrib['name']
            # Go out and find this jawns
            imp_block = _find_import(block.attrib['src'], infile)
            imp_path = os.path.split(block.attrib['src'])[0]
            parse_block(imp_block, make_unique_block_id(), imp_name=imp_name, imp_path=imp_path)
        else:
            parse_block(block, make_unique_block_id())

    # Root model exports - their names might have changed!
    for child in root.findall("export"):

        child_type = DEFAULT_TYPE
        if 'type' in child.attrib:
            child_type = child.attrib['type']

        naive_name = _resolve_attr(child.attrib['from'], bindings, expect_value=False)

        if naive_name in name_resolves:
            name = name_resolves[naive_name]
        else:
            name = naive_name

        child_type = _resolve_attr(child_type, bindings, expect_value=False)
        dim = _resolve_attr(child.attrib['dim'], bindings, expect_value=True)
        x = onnx.helper.make_tensor_value_info(name,
                                            ONNX_TYPE_DICT[child_type],
                                            dim)
        all_outputs.append(x)

    # Create the graph (GraphProto)
    graph_def = onnx.helper.make_graph(
        nodes=all_nodes,
        name=graph_name,
        inputs=all_inputs,  # Graph input
        outputs=all_outputs,  # Graph output
        initializer=all_inits,
    )

    # Create the model (ModelProto)
    model_def = onnx.helper.make_model(graph_def, producer_name=producer)
    model_def.opset_import[0].version = 13

    model_def = onnx.shape_inference.infer_shapes(model_def)

    onnx.checker.check_model(model_def)

    onnx.save(model_def, _get_output_model_name(outfile, infile))

    # Write the weights file if we're supposed to
    if write_weights:
        weights_path = os.path.join(infile, WEIGHTS_FNAME)
        with open(weights_path, "wb") as fhand:
            pickle.dump(weights, fhand)



__all__ = [
    'OnnxBatchNorm',
]

import torch
import torch.nn.functional as F
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OnnxMapping
from agrippa.onnx2torch.utils.common import OnnxToTorchModule
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import get_shape_from_value_info
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node

_BN_CLASS_FROM_SPATIAL_RANK = {
    0: nn.BatchNorm1d,
    1: nn.BatchNorm1d,
    2: nn.BatchNorm2d,
    3: nn.BatchNorm3d,
}


class OnnxBatchNorm(nn.Module, OnnxToTorchModule):  # pylint: disable=missing-docstring
    def __init__(self, momentum: float, epsilon: float):
        super().__init__()
        self.momentum = momentum
        self.epsilon = epsilon

    def forward(  # pylint: disable=missing-function-docstring
        self,
        input_data: torch.Tensor,
        weight: torch.Tensor,
        bias: torch.Tensor,
        running_mean: torch.Tensor,
        running_var: torch.Tensor,
    ) -> torch.Tensor:
        return F.batch_norm(
            input_data,
            running_mean=running_mean,
            running_var=running_var,
            weight=weight,
            bias=bias,
            training=self.training,
            momentum=self.momentum,
            eps=self.epsilon,
        )


@add_converter(operation_type='BatchNormalization', version=15)
@add_converter(operation_type='BatchNormalization', version=14)
@add_converter(operation_type='BatchNormalization', version=9)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:
    node_attributes = node.attributes
    epsilon = node_attributes.get('epsilon', 1e-5)
    momentum = 1 - node_attributes.get('momentum', 0.9)  # See PyTorch documentation for batch norm.

    if all(value_name in graph.initializers for value_name in node.input_values[1:]) and len(node.output_values) == 1:
        input_value_info = graph.value_info[node.input_values[0]]
        input_shape = get_shape_from_value_info(input_value_info)
        spatial_rank = len(input_shape) - 2
        try:
            bn_class = _BN_CLASS_FROM_SPATIAL_RANK[spatial_rank]
        except KeyError as exc:
            raise NotImplementedError(
                f'BatchNorm operation with spatial rank == {spatial_rank} is not implemented'
            ) from exc

        scale_value_name = node.input_values[1]
        bias_value_name = node.input_values[2]
        mean_value_name = node.input_values[3]
        var_value_name = node.input_values[4]

        scale = graph.initializers[scale_value_name].to_torch()
        torch_module = bn_class(
            num_features=scale.size()[0],
            eps=epsilon,
            momentum=momentum,
        )
        with torch.no_grad():
            torch_module.running_mean.data = graph.initializers[mean_value_name].to_torch()
            torch_module.running_var.data = graph.initializers[var_value_name].to_torch()
            torch_module.weight.data = scale
            torch_module.bias.data = graph.initializers[bias_value_name].to_torch()

        onnx_mapping = OnnxMapping(
            inputs=(node.input_values[0],),
            outputs=node.output_values,
        )
    else:
        if len(node.output_values) != 1:
            raise NotImplementedError('BatchNorm operation with mean/var output is not implemented')

        torch_module = OnnxBatchNorm(momentum=momentum, epsilon=epsilon)
        onnx_mapping = onnx_mapping_from_node(node)

    return OperationConverterResult(
        torch_module=torch_module,
        onnx_mapping=onnx_mapping,
    )



__all__ = [
    'OnnxNonMaxSuppression',
]

from typing import Optional

import torch
import torch._C as torch_C
import torchvision
from torch import nn

from agrippa.onnx2torch.node_converters.registry import add_converter
from agrippa.onnx2torch.onnx_graph import OnnxGraph
from agrippa.onnx2torch.onnx_node import OnnxNode
from agrippa.onnx2torch.utils.common import OperationConverterResult
from agrippa.onnx2torch.utils.common import onnx_mapping_from_node
from agrippa.onnx2torch.utils.custom_export_to_onnx import CustomExportToOnnx
from agrippa.onnx2torch.utils.custom_export_to_onnx import OnnxToTorchModuleWithCustomExport


class OnnxNonMaxSuppression(nn.Module, OnnxToTorchModuleWithCustomExport):  # pylint: disable=missing-class-docstring
    def __init__(self, center_point_box: bool = False):
        super().__init__()
        self.center_point_box = center_point_box

    def _do_forward(
        self,
        boxes: torch.Tensor,
        scores: torch.Tensor,
        max_output_boxes_per_class: Optional[torch.Tensor],
        iou_threshold: Optional[torch.Tensor],
        score_threshold: Optional[torch.Tensor],
    ) -> torch.Tensor:
        if max_output_boxes_per_class is None:
            return torch.empty([0, 3], dtype=torch.int64, device=boxes.device)

        max_output_boxes_per_class = max_output_boxes_per_class.item()
        iou_threshold = 0.0 if iou_threshold is None else iou_threshold.item()
        score_threshold = 0.0 if score_threshold is None else score_threshold.item()

        out = []
        # boxes - [bs, num_boxes, 4], scores - [bs, n_classes, num_boxes]
        for batch_index, (batch_boxes, batch_scores) in enumerate(zip(boxes, scores)):
            # bbox - [num_boxes, 4], score - [n_classes, num_boxes]
            for class_index, class_scores in enumerate(batch_scores):
                confidence_mask = class_scores > score_threshold
                confidence_indexes = confidence_mask.nonzero(as_tuple=False).squeeze(1)

                filtered_batch_boxes = batch_boxes[confidence_indexes]
                if self.center_point_box:
                    filtered_batch_boxes = torchvision.ops.box_convert(
                        filtered_batch_boxes,
                        in_fmt='cxcywh',
                        out_fmt='xyxy',
                    )

                nms_indexes = torchvision.ops.nms(
                    boxes=filtered_batch_boxes,
                    scores=class_scores[confidence_indexes],
                    iou_threshold=iou_threshold,
                )
                num_boxes = min(max_output_boxes_per_class, nms_indexes.size(0))
                nms_indexes = nms_indexes[:num_boxes]
                indexes = confidence_indexes[nms_indexes]

                out.extend([batch_index, class_index, box_index] for box_index in indexes)
        if len(out) == 0:
            return torch.empty([0, 3], dtype=torch.int64, device=boxes.device)

        return torch.tensor(out, dtype=torch.int64, device=boxes.device)

    def forward(  # pylint: disable=missing-function-docstring
        self,
        boxes: torch.Tensor,
        scores: torch.Tensor,
        max_output_boxes_per_class: Optional[torch.Tensor] = None,
        iou_threshold: Optional[torch.Tensor] = None,
        score_threshold: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        output = self._do_forward(boxes, scores, max_output_boxes_per_class, iou_threshold, score_threshold)
        if torch.onnx.is_in_onnx_export():
            if max_output_boxes_per_class is None:
                max_output_boxes_per_class = torch.tensor([0], dtype=torch.int64)
            if iou_threshold is None:
                iou_threshold = torch.tensor([0.0], dtype=torch.float32)
            if score_threshold is None:
                score_threshold = torch.tensor([0.0], dtype=torch.float32)

            return _NmsExportToOnnx.set_output_and_apply(
                output,
                boxes,
                scores,
                max_output_boxes_per_class,
                iou_threshold,
                score_threshold,
                int(self.center_point_box),
            )

        return output


class _NmsExportToOnnx(CustomExportToOnnx):  # pylint: disable=abstract-method
    @staticmethod
    def symbolic(graph: torch_C.Graph, *args) -> torch_C.Value:
        boxes, scores, max_output_boxes_per_class, iou_threshold, score_threshold, center_point_box = args
        return graph.op(
            'NonMaxSuppression',
            boxes,
            scores,
            max_output_boxes_per_class,
            iou_threshold,
            score_threshold,
            center_point_box_i=center_point_box,
            outputs=1,
        )


@add_converter(operation_type='NonMaxSuppression', version=10)
@add_converter(operation_type='NonMaxSuppression', version=11)
def _(node: OnnxNode, graph: OnnxGraph) -> OperationConverterResult:  # pylint: disable=unused-argument
    return OperationConverterResult(
        torch_module=OnnxNonMaxSuppression(center_point_box=node.attributes.get('center_point_box', 0) == 1),
        onnx_mapping=onnx_mapping_from_node(node),
    )
